import logging
import pandas as pd
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import confusion_matrix
from omniscient.utils.constants import *
from omniscient.utils.file_handler import FileHandler
from deeppavlov.core.common.chainer import Chainer
from typing import List


class Predictor:
    """
    Manipulating data frames and using an initialised model to generate predictions
    """
    def __init__(self, model: Chainer)->None:
        """
        Sets variable model to initialised model object passed in
        :param model: An initialised model
        """
        logging.info("Initialising Predictor")
        self.model = model
        self.file_handler = FileHandler()

    def predict_instance(self, context: str, question: str)->str:
        """
        Returns matched snippet for context and question
        :param context: Context (e.g. document) about which question is asked
        :param question: Question asked
        :return: Matched snippet
        """
        return self.model([context], [question])[0][0]

    def predict_from_file(self, input_file: str, output_file: str)->None:
        """
        Predict labels for each row in a file with the question and sentence on the same row
        :param input_file: File with question and sentence on same row
        :param output_file: File to output DataFrame with Predicted Labels
        """
        try:
            df = self.file_handler.read_dataframe_from_input_file(input_file)
            df_predictions = self.predict_batch(df)  # contains context column
            df_predictions = df_predictions.drop(columns='Context')

            self.file_handler.write_dataframe_to_output_file(df_predictions, output_file)

        except IOError as e:
            logging.error(e)


    def predict_batch(self, df: pd.DataFrame)->pd.DataFrame:
        """
        Predicting on batch of instances (e.g. as converted from a file)
        :param df: Dataframe with QuestionID, Question, Context etc.
        :return: Dataframe with PredictedLabel column attached
        """
        logging.info("Predicting on batch of instances")
        df_contexts = self._create_contexts_dataframe(df)

        # New DataFrame to hold QuestionID and the corresponding predicted answer
        df_qid_ans = pd.DataFrame(columns=['QuestionID', 'PredictedAnswer'])
        for index, row in df_contexts.iterrows():
            # NOTE: pred_ans is the snippet which has the answer, not the whole sentence
            pred_ans = self.predict_instance(row['Context'], row['Question'])
            df_qid_ans = df_qid_ans.append({'QuestionID': row['QuestionID'], 'PredictedAnswer': pred_ans},
                                           ignore_index=True)
            logging.debug("QuestionID: {} Question: {} Predicted Snippet: {}"
                          .format(row['QuestionID'], row['Question'], pred_ans))

        df_with_pred_ans = pd.merge(df, df_qid_ans, on='QuestionID')

        predictions = self._get_prediction_labels(df_with_pred_ans)
        df['PredictedLabel'] = predictions

        logging.info("Precision, Recall, F-Score, Support: {}"
                     .format(precision_recall_fscore_support(df['Label'], df['PredictedLabel'])))
        logging.info("Confusion Matrix: {}".format(confusion_matrix(df['Label'], df['PredictedLabel'])))

        return df

    def _create_contexts_dataframe(self, df:pd.DataFrame)->None:
        """
        Add 'Context' column to the raw dataframe which represents the concatenateion of all the sentences in the file
        :param df: Raw dataframe
        :return: Raw dataframe with 'Context' column added
        """
        # Add new 'context' column which combines all sentences relevant to a Question (to resemble a document)
        df['Context'] = pd.DataFrame(df.groupby(['QuestionID'])['Sentence'].transform(lambda x: ' '.join(x).lower()))

        # Create a new DataFrame with just the QuestionID, Question and the combined text. There will be duplicates
        # equal to the number of sentences - drop these as well
        df_contexts = df[['QuestionID', 'Question', 'Context']].drop_duplicates()

        return df_contexts

    def _get_prediction_labels(self, df_with_pred_ans: pd.DataFrame)->List:
        """
        Check for match of answer snippet string with Sentence for each sentence so that each sentence can be labeled
        as 0 or 1

        :param df_with_pred_ans: DataFrame with 'PredictedAnswer' column
        :return: for each row in DataFrame whether the string in the 'PredictedAnswer' column is found in the sentence
        """
        predictions = []
        for index, row in df_with_pred_ans.iterrows():
            if row['PredictedAnswer'].lower() in row['Sentence'].lower():
                predictions.append(1)
            else:
                predictions.append(0)

        return predictions






from unittest import TestCase
from deeppavlov import build_model, configs
from omniscient.prediction.predictor import Predictor
import pandas as pd


class TestModel(TestCase):
    def setUp(self):
        self.model = build_model(configs.squad.squad, download=True)
        self.predictor = Predictor(self.model)

    def test_model_returns_correct_response_for_single_instance(self):
        ans = self.predictor.predict_instance('DeepPavlov is library for NLP and dialog systems.',
                                              'What is DeepPavlov?')

        assert ans == 'library for NLP and dialog systems'

    def test_model_returns_correct_response_for_batch(self):
        df = pd.DataFrame({'QuestionID': ['Q1', 'Q1', 'Q1', 'Q1', 'Q1'],
                           'Question': ['how are glacier caves formed?', 'how are glacier caves formed?',
                                        'how are glacier caves formed?', 'how are glacier caves formed?',
                                        'how are glacier caves formed?'],
                           'DocumentID': ['D1', 'D1', 'D1', 'D1', 'D1'],
                           'SentenceID': ['D1-0', 'D1-1', 'D1-2', 'D1-3', 'D1-4'],
                           'Sentence': ['A partly submerged glacier cave on Perito Moreno Glacier .',
                                        'The ice facade is approximately 60 m high',
                                        'Ice formations in the Titlis glacier cave',
                                        'A glacier cave is a cave formed within the ice of a glacier .',
                                        'Glacier caves are often called ice caves , but this term is properly used to '
                                        'describe bedrock caves that contain year-round ice.'],
                           'Label': [0, 0, 0, 1,0 ]
                           })
        df_ans = self.predictor.predict_batch(df)

        ans = df_ans[df_ans['PredictedLabel'] == 1]['Sentence'].tolist()[0]

        assert ans == 'A glacier cave is a cave formed within the ice of a glacier .'



from deeppavlov import build_model, configs
from omniscient.endpoint.rest_endpoint import RestEndpoint
from deeppavlov.core.common.chainer import Chainer
import logging


logger = logging.getLogger(__name__)


def init_model()->Chainer:
    """
    Initialises model. Downloads pretrained weights if necessary (hash check conducted before download)
    :return: The initialised model that can be passed to a Predictor object
    """
    logging.info("Initialising SQuAD model")
    model = build_model(configs.squad.squad, download=True)

    return model


def init_endpoint(model: Chainer, host: str='0.0.0.0', port: int=5000):
    """
    Initialises an REST endpoint to query for predictions
    :param model: The model initialised via init_model()
    :param host: Host address to run endpoint. NOTE '0.0.0.0' for externally visible server
    :param port: Port to run endpoint on
    :return: The initialised REST endpoint
    """
    restendpoint = RestEndpoint(model, host, port)


if __name__ == '__main__':
    # If run as main will initialise model and an endpoint that uses that model to provide predictions
    logging.basicConfig(level=logging.DEBUG)

    mdl = init_model()
    init_endpoint(mdl)



from setuptools import setup, find_packages

REQUIRED_PACKAGES = [
    'scikit-learn',
    'pandas',
    'deeppavlov',
    'tensorflow',
    'flask'
]


setup(name='omniscient',
      version='0.7.4',
      description='Omniscient question answering system',
      long_description='Omniscient Question Answering System. Uses a BERT model fine-tuned on the SQuAD dataset.\
Returns predictions based on a Context and a Question. ',
      url='https://github.com/pasank/omniscient',
      author='Pasan Karunaratne',
      author_email='pasankarunaratne@gmail.com',
      license='MIT',
      packages=['omniscient', 'omniscient/endpoint', 'omniscient/configs', 'omniscient/prediction', 'omniscient/tests',
                'omniscient/utils'],  # TODO MANIFEST
      install_requires=REQUIRED_PACKAGES,
      zip_safe=False,
      )



from flask import Flask, request
from omniscient.prediction.predictor import Predictor
import logging
from deeppavlov.core.common.chainer import Chainer
from omniscient.utils.constants import *


class RestEndpoint:
    """
    REST endpoint that enables question answering using an initialised model
    """

    def __init__(self, model: Chainer, host: str='0.0.0.0', port: int=5000)->None:
        """
        Initialises REST endpoint. Root will be a note on how to use the endpoint.
        /predict will be a POST endpoint which requires the specification of values for
         'context' and 'question' keys
        :param model: An initalised model
        :param host: Host address to run endpoint. NOTE '0.0.0.0' for externally visible server
        :param port: Port to run endpoint on
        """
        self.model = model
        self.predictor = Predictor(model)

        logging.info("Initialising REST endpoint using Flask")

        self.app = Flask(__name__)
        self.app.add_url_rule('/', 'index', self.index_action)
        self.app.add_url_rule('/predict', 'predict', self.predict_action, methods=['POST'])

        self.app.run(host=host, port=port)

    def predict_action(self)->str:
        """
        For the provided 'context' and 'question' in the /predict POST request, returns the
        matched snippet in the context.
        :return: Returns the matched snippet in the context for the question provided
        """
        data = request.get_json(force=True)

        context = data['context']
        question = data['question']

        logging.info("Received context: {}".format(context))
        logging.info("Received question: {}".format(question))

        matched_snippet = self.predictor.predict_instance(context, question)

        logging.info("Matched answer snippet: {}".format(matched_snippet))

        return matched_snippet

    def index_action(self)->str:
        """
        Returns a usage text
        :return: Text on how to use Omniscient REST endpoint for prediction
        """
        return "Omniscient Question Answering System. POST to /predict with 'context' and 'question' "



from unittest import TestCase
from deeppavlov import build_model, configs
import pytest


class TestModel(TestCase):
    def setUp(self):
        self.model = build_model(configs.squad.squad, download=True)

    def test_model_returns_correct_response_for_single_instance(self):
        ans = self.model(['DeepPavlov is library for NLP and dialog systems.'], ['What is DeepPavlov?'])[0][0]

        assert ans == 'library for NLP and dialog systems'



import pandas as pd
from omniscient.utils.constants import *
import logging


class FileHandler:
    """
    Reading and writing CSV files to/from Pandas DataFrames
    """
    def __init__(self)->None:
        """
        Initialising FileHandler
        """
        logging.info("Initialising FileReader")

    def read_dataframe_from_input_file(self, input_file: str)->pd.DataFrame:
        """
        Read raw DataFrame from input CSV file
        :param input_file: input CSV file
        :return: Pandas DataFrame correspodning to CSV file
        """
        try:
            logging.info("Reading dataframe from: {}".format(input_file))
            df = pd.read_csv(input_file, sep=INPUT_FILE_SEPARATOR)

            return df
        except IOError as e:
            raise

    def write_dataframe_to_output_file(self, df: pd.DataFrame, output_file: str)->None:
        """
        Write DataFrame to output CSV file
        :param df: input DataFrame (commonly one with the PredictedLabel column added)
        :param output_file: File to write out DataFrame to
        :return:
        """
        try:
            logging.info("Writing dataframe to: {}".format(output_file))
            df.to_csv(output_file)
        except IOError as e:
            raise